# pylint: disable=R0913,W0718,W0706,C0415
#  Drakkar-Software OctoBot-Commons
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import copy
import os
import uuid

import octobot_commons.profiles.profile_data as profile_data_import
import octobot_commons.profiles.profile as profile_import
import octobot_commons.logging as bot_logging
import octobot_commons.json_util as json_util
import octobot_commons.constants as constants
import octobot_commons.aiohttp_util as aiohttp_util
import octobot_commons.enums as enums

IMPORTED_AVATAR = "avatar"
IMPORTED_PROFILES_DEFAULT_EXTRA_BACKTESTING_TIMEFRAME = (
    enums.TimeFrames.FIFTEEN_MINUTES.value
)


def init_profile_directory(
    output_path: str,
):
    """
    :param output_path: profile folder path
    """
    if os.path.exists(output_path):
        raise OSError(f"{output_path} already exists")
    os.mkdir(output_path)


async def convert_profile_data_to_profile_directory(
    profile_data: profile_data_import.ProfileData,
    output_path: str,
    description: str = None,
    risk: enums.ProfileRisk = enums.ProfileRisk.MODERATE,
    auto_update: bool = False,
    slug: str = None,
    avatar_url: str = None,
    force_simulator: bool = False,
    aiohttp_session=None,
    profile_to_update: profile_import.Profile = None,
    changed: bool = False,
) -> bool:
    """
    Creates a profile folder from the given ProfileData
    :param profile_data: path to the profile zipped archive
    :param description: profile description
    :param risk: profile risk
    :param slug: slug of the associated strategy
    :param auto_update: True if the profile should be kept up-to-date
    :param avatar_url: profile avatar_url
    :param output_path: profile folder path
    :param force_simulator: True if trader simulator should be forced in config
    :param aiohttp_session: session to use
    :param profile_to_update: profile to update instead of creating a new one
    :param changed: if True, profile will be saved even if no change are identified
    """
    profile = (
        profile_to_update
        if profile_to_update
        else _get_profile(
            profile_data,
            description,
            risk,
            output_path,
            auto_update,
            slug,
            force_simulator,
        )
    )
    # when updating profile, keep existing registered tentacles
    import_registered_tentacles = profile_to_update is not None
    # tentacles_config.json
    tentacles_setup_config = _get_tentacles_setup_config(
        profile_data, output_path, import_registered_tentacles
    )
    if tentacles_setup_config.save_config(is_config_update=True):
        changed = True
    # specific_config
    if _save_specific_config(profile_data, output_path, bool(profile_to_update)):
        changed = True
    # avatar file
    if avatar_url:
        try:
            await _download_and_set_avatar(
                profile, avatar_url, output_path, aiohttp_session
            )
        except Exception as err:
            bot_logging.get_logger(__name__).exception(
                err, True, f"Error when downloading profile avatar: {err}"
            )
    # finish with profile.json to include edits from previous methods
    if changed:
        profile.save()
    return changed


def _get_profile(
    profile_data: profile_data_import.ProfileData,
    description: str,
    risk: enums.ProfileRisk,
    output_path: str,
    auto_update: bool,
    slug: str,
    force_simulator: bool,
):
    profile = profile_data.to_profile(output_path)
    if force_simulator:
        profile.config[constants.CONFIG_TRADER][constants.CONFIG_ENABLED_OPTION] = False
        profile.config[constants.CONFIG_SIMULATOR][
            constants.CONFIG_ENABLED_OPTION
        ] = True
    profile.description = description
    profile.risk = risk
    profile.auto_update = auto_update
    profile.slug = slug
    profile.profile_id = str(uuid.uuid4().hex)
    profile.read_only = True
    profile.extra_backtesting_time_frames = [
        IMPORTED_PROFILES_DEFAULT_EXTRA_BACKTESTING_TIMEFRAME
    ]
    return profile


def get_updated_profile(
    profile_to_update: profile_import.Profile,
    profile_data: profile_data_import.ProfileData,
) -> bool:
    """
    :param profile_to_update: the profile to be updated
    :param profile_data: the profile_data to get the update from
    :return: True if something changed in the updated profile
    """
    updated_profile = profile_data.to_profile("")
    changed = False
    # update traded currencies (add new currencies)
    origin_currencies = copy.deepcopy(
        profile_to_update.config[constants.CONFIG_CRYPTO_CURRENCIES]
    )
    profile_to_update.config[constants.CONFIG_CRYPTO_CURRENCIES] = {
        **origin_currencies,
        **updated_profile.config[constants.CONFIG_CRYPTO_CURRENCIES],
    }
    if (
        origin_currencies
        != profile_to_update.config[constants.CONFIG_CRYPTO_CURRENCIES]
    ):
        changed = True
    # update ref market
    origin_ref_market = profile_to_update.config[constants.CONFIG_TRADING][
        constants.CONFIG_TRADER_REFERENCE_MARKET
    ]
    profile_to_update.config[constants.CONFIG_TRADING][
        constants.CONFIG_TRADER_REFERENCE_MARKET
    ] = profile_data.trading.reference_market
    if (
        origin_ref_market
        != profile_to_update.config[constants.CONFIG_TRADING][
            constants.CONFIG_TRADER_REFERENCE_MARKET
        ]
    ):
        changed = True
    # leave other fields as is (tentacles config will be updated)
    return changed


def _get_tentacles_setup_config(
    profile_data: profile_data_import.ProfileData,
    output_path: str,
    import_registered_tentacles: bool,
):
    try:
        import octobot_tentacles_manager.api

        classes = [
            octobot_tentacles_manager.api.get_tentacle_class_from_string(
                tentacle_data.name
            ).get_name()
            for tentacle_data in profile_data.tentacles
        ]
        config_path = os.path.join(output_path, constants.CONFIG_TENTACLES_FILE)
        tentacles_setup_config = (
            octobot_tentacles_manager.api.create_tentacles_setup_config_with_tentacles(
                *classes, config_path=config_path
            )
        )
        use_reference_registered_tentacles = (
            not tentacles_setup_config.registered_tentacles
        )
        octobot_tentacles_manager.api.fill_with_installed_tentacles(
            tentacles_setup_config,
            import_registered_tentacles=import_registered_tentacles,
            use_reference_registered_tentacles=use_reference_registered_tentacles,
        )
        return tentacles_setup_config
    except ImportError:
        raise


def _save_specific_config(
    profile_data: profile_data_import.ProfileData,
    output_path: str,
    is_config_update: bool,
) -> bool:
    changed = False
    try:
        import octobot_tentacles_manager.constants

        specific_config_dir = os.path.join(
            output_path,
            octobot_tentacles_manager.constants.TENTACLES_SPECIFIC_CONFIG_FOLDER,
        )
        if not os.path.exists(specific_config_dir):
            os.mkdir(specific_config_dir)
        for tentacle_config in profile_data.tentacles:
            file_path = os.path.join(
                specific_config_dir,
                f"{tentacle_config.name}{octobot_tentacles_manager.constants.CONFIG_EXT}",
            )
            if is_config_update and json_util.has_same_content(
                file_path, tentacle_config.config
            ):
                # nothing to do
                continue
            changed = True
            json_util.safe_dump(
                tentacle_config.config,
                file_path,
            )
    except ImportError:
        raise
    return changed


async def _download_and_set_avatar(
    profile, avatar_url, output_path: str, aiohttp_session
):
    profile.avatar = _get_avatar_name(avatar_url)
    try:
        import aiofiles

        async with aiofiles.open(
            os.path.join(output_path, profile.avatar), "wb+"
        ) as downloaded_file:
            await aiohttp_util.download_stream_file(
                downloaded_file,
                avatar_url,
                aiohttp_session,
                is_aiofiles_output_file=True,
            )
    except ImportError:
        raise


def _get_avatar_name(avatar_url: str):
    # remove params and get file name with ext
    return avatar_url.split("?")[0].split("/")[-1]



# pylint: disable=C0103,R0902
#  Drakkar-Software OctoBot-Commons
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import dataclasses

import octobot_commons.dataclasses
import octobot_commons.constants


@dataclasses.dataclass
class ExchangeAuthData(octobot_commons.dataclasses.FlexibleDataclass):
    internal_name: str
    exchange_credential_id: str = ""
    api_key: str = ""
    api_secret: str = ""
    api_password: str = ""
    exchange_type: str = octobot_commons.constants.DEFAULT_EXCHANGE_TYPE
    sandboxed: bool = False

    def apply_to_exchange_config(self, config):
        """
        Updates the given Configuration object to use the local authentication data
        :param config: Configuration object to update
        """
        applied = False
        for exchange, exchange_config in config.config[
            octobot_commons.constants.CONFIG_EXCHANGES
        ].items():
            if exchange == self.internal_name:
                self._apply_config(exchange_config)
                applied = True
                break
        if not applied:
            # exchange doesn't already exist: add it
            exchange_config = {octobot_commons.constants.CONFIG_ENABLED_OPTION: True}
            self._apply_config(exchange_config)
            config.config[octobot_commons.constants.CONFIG_EXCHANGES][
                self.internal_name
            ] = exchange_config

    def _apply_config(self, exchange_config: dict):
        exchange_config[octobot_commons.constants.CONFIG_EXCHANGE_KEY] = self.api_key
        exchange_config[
            octobot_commons.constants.CONFIG_EXCHANGE_SECRET
        ] = self.api_secret
        exchange_config[
            octobot_commons.constants.CONFIG_EXCHANGE_PASSWORD
        ] = self.api_password
        exchange_config[
            octobot_commons.constants.CONFIG_EXCHANGE_SANDBOXED
        ] = self.sandboxed
        exchange_config[
            octobot_commons.constants.CONFIG_EXCHANGE_TYPE
        ] = self.exchange_type



#  Drakkar-Software OctoBot-Commons
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.


class ConfigError(Exception):
    """
    Config related Exception
    """


class RemoteConfigError(ConfigError):
    """
    Fetched config related Exception
    """


class NoProfileError(Exception):
    """
    Profile related Exception: raised when the current profile can't be found and default profile can't be loaded
    """


class ProfileConflictError(Exception):
    """
    Profile related Exception: raised when the current profile can't be renamed as expected
    """


class ProfileRemovalError(Exception):
    """
    Profile related Exception: raised when the current profile can't be can't be removed
    """


class ProfileImportError(Exception):
    """
    Profile related Exception: raised when the imported profile is invalid
    """


class ConfigEvaluatorError(Exception):
    """
    Evaluator config related Exception
    """


class ConfigTradingError(Exception):
    """
    Trading config related Exception
    """


class TentacleNotFound(Exception):
    """
    Tentacle not found related Exception
    """


class UninitializedCache(Exception):
    """
    Raised when a cache is requested but has not yet been initialized
    """


class NoCacheValue(Exception):
    """
    Raised when a cache value is selected but is not available in database
    """


class UncachableValue(Exception):
    """
    Raised when a cache value is selected but is not available in database
    """


class DatabaseNotFoundError(Exception):
    """
    Raised when a database can't be found
    """


class MissingDataError(Exception):
    """
    Raised when there is not enough available candles
    """


class MissingExchangeDataError(Exception):
    """
    Raised when there is no available data for this exchange
    """


class ExecutionAborted(Exception):
    """
    Raised when the current execution should be aborted
    """


class LogicalOperatorError(Exception):
    """
    Raised when a logical operation is invalid
    """


class UnsupportedError(Exception):
    """
    Raised when an unsupported message is received
    """


class InvalidUserInputError(Exception):
    """
    Raised when a user input in invalid
    """


class MissingSignalBuilder(Exception):
    """
    Raised when a signal builder is not found
    """



# pylint: disable=W1203,R0902,R0913,R0914
#  Drakkar-Software OctoBot-Commons
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import contextlib

import octobot_commons.enums as enums
import octobot_commons.dict_util as dict_util


class UserInput:
    MAX_ORDER = 9999

    def __init__(
        self,
        name,
        input_type,
        value,
        def_val,
        tentacle_type,
        tentacle_name,
        min_val=None,
        max_val=None,
        options=None,
        title=None,
        item_title=None,
        other_schema_values=None,
        editor_options=None,
        read_only=False,
        is_nested_config=None,
        nested_tentacle=None,
        parent_input_name=None,
        show_in_summary=True,
        show_in_optimizer=True,
        path=None,
        order=None,
    ):
        self.name = name
        self.input_type = input_type
        self.value = value
        self.def_val = def_val
        self.tentacle_type = tentacle_type
        self.tentacle_name = tentacle_name
        self.min_val = min_val
        self.max_val = max_val
        self.options = options
        self.title = title
        self.item_title = item_title
        self.other_schema_values = other_schema_values
        self.editor_options = editor_options
        self.read_only = read_only
        self.is_nested_config = is_nested_config
        self.nested_tentacle = nested_tentacle
        self.parent_input_name = parent_input_name
        self.show_in_summary = show_in_summary
        self.show_in_optimizer = show_in_optimizer
        self.path = path
        self.order = order

    def to_dict(self):
        """
        :return: the dict representation of the UserInput
        """
        return {
            "name": self.name,
            "input_type": self.input_type
            if isinstance(self.input_type, str)
            else self.input_type.value,
            "value": self.value,
            "def_val": self.def_val,
            "min_val": self.min_val,
            "max_val": self.max_val,
            "options": self.options,
            "title": self.title,
            "item_title": self.item_title,
            "other_schema_values": self.other_schema_values,
            "editor_options": self.editor_options,
            "read_only": self.read_only,
            "tentacle_type": self.tentacle_type,
            "tentacle": self.tentacle_name,
            "nested_tentacle": self.nested_tentacle,
            "parent_input_name": self.parent_input_name,
            "is_nested_config": self.is_nested_config,
            "in_summary": self.show_in_summary,
            "in_optimizer": self.show_in_optimizer,
            "path": self.path,
            "order": self.order,
        }


class UserInputFactory:
    def __init__(self, user_input_tentacle_type: enums.UserInputTentacleTypes):
        self.user_input_tentacle_type: enums.UserInputTentacleTypes = (
            user_input_tentacle_type
        )
        self.tentacle_class = None
        self.tentacle_config_proxy = None

    def set_tentacle_class(self, tentacle_class):
        """
        set the associated tentacle class
        :return: self
        """
        self.tentacle_class = tentacle_class
        return self

    def set_tentacle_config_proxy(self, tentacle_config_proxy):
        """
        set the associated tentacle configuration proxy function
        :return: self
        """
        self.tentacle_config_proxy = tentacle_config_proxy
        return self

    def user_input(
        self,
        name: str,
        input_type,
        def_val,
        registered_inputs: dict,
        value=None,
        min_val=None,
        max_val=None,
        options=None,
        title=None,
        item_title=None,
        other_schema_values=None,
        editor_options=None,
        read_only=False,
        is_nested_config=None,
        nested_tentacle=None,
        parent_input_name=None,
        show_in_summary=True,
        show_in_optimizer=True,
        path=None,
        order=None,
        array_indexes=None,
        return_value_only=True,
        update_parent_value=True,
    ):
        """
        Set and return a user input value.
        The returned value is set as an attribute named as the "name" param with " " replaced by "_"
        in self.specific_config.
        Types are any UserInputTypes
        :return: the saved_config value if any, def_val otherwise
        """
        sanitized_name = sanitize_user_input_name(name)
        parent = _find_parent_config_node(
            self.tentacle_config_proxy(), parent_input_name, array_indexes
        )
        used_value = value
        if value is None:
            # value is not provided, use def_val
            used_value = def_val
            if parent is not None:
                # parent found, try to find saved value
                try:
                    used_value = parent[sanitized_name]
                except KeyError:
                    # use default value
                    pass
        input_key = f"{parent_input_name}{name}"
        created_input = UserInput(
            name,
            input_type,
            used_value,
            def_val,
            self.user_input_tentacle_type.value,
            self.tentacle_class.get_name(),
            min_val=min_val,
            max_val=max_val,
            options=options,
            title=title,
            item_title=item_title,
            other_schema_values=other_schema_values,
            editor_options=editor_options,
            read_only=read_only,
            is_nested_config=is_nested_config,
            nested_tentacle=nested_tentacle,
            parent_input_name=parent_input_name,
            show_in_summary=show_in_summary,
            show_in_optimizer=show_in_optimizer,
            path=path,
            order=order,
        )
        if input_key not in registered_inputs:
            # do not register user input multiple times
            registered_inputs[input_key] = created_input
        if parent is not None and update_parent_value:
            parent[sanitized_name] = used_value
        return used_value if return_value_only else created_input

    @contextlib.contextmanager
    def local_factory(self, tentacle_class, tentacle_config_proxy):
        """
        temporarily set the associated tentacle class and tentacle config proxy
        """
        previous_tentacle_class = self.tentacle_class
        previous_tentacle_config_proxy = self.tentacle_config_proxy
        try:
            self.set_tentacle_class(tentacle_class).set_tentacle_config_proxy(
                tentacle_config_proxy
            )
            yield
        finally:
            self.set_tentacle_class(previous_tentacle_class).set_tentacle_config_proxy(
                previous_tentacle_config_proxy
            )


def sanitize_user_input_name(name):
    """
    :return: the sanitized user input name
    """
    return name.replace(" ", "_")


async def save_user_input(
    u_input: UserInput,
    run_data_writer,
    flush_if_necessary=False,
    skip_flush=False,
):
    """
    Save the user input in the given run_data_writer. First checks if the user input is not already present.
    Does not update a user input if it is already saved in run_data_writer.
    """
    if not run_data_writer.enable_storage:
        return
    if not await run_data_writer.contains_row(
        enums.DBTables.INPUTS.value,
        {
            "name": u_input.name,
            "tentacle": u_input.tentacle_name,
            "nested_tentacle": u_input.nested_tentacle,
            "parent_input_name": u_input.parent_input_name,
            "is_nested_config": u_input.is_nested_config,
        },
    ):
        await run_data_writer.log(
            enums.DBTables.INPUTS.value,
            u_input.to_dict(),
        )
        if not skip_flush and (
            flush_if_necessary or run_data_writer.are_data_initialized
        ):
            # in some cases, user inputs might be setup after the 1st trading mode cycle: flush
            # writer in live mode to ensure writing
            await run_data_writer.flush()


def get_user_input_tentacle_type(tentacle) -> str:
    """
    :return: the tentacle associated UserInputTentacleTypes
    """
    return (
        enums.UserInputTentacleTypes.TRADING_MODE.value
        if hasattr(tentacle, "trading_config")
        else enums.UserInputTentacleTypes.EVALUATOR.value
        if hasattr(tentacle, "specific_config")
        else enums.UserInputTentacleTypes.EXCHANGE.value
    )


async def get_user_inputs(reader, tentacle_name=None):
    """
    :return: all user inputs. Only user inputs associated to the given tentacle_name that have been saved into
    the given reader if tentacle_name is given
    """
    all_inputs = await reader.all(enums.DBTables.INPUTS.value)
    if tentacle_name is None:
        return all_inputs
    return [
        selected_input
        for selected_input in all_inputs
        if selected_input["tentacle"] == tentacle_name
    ]


async def clear_user_inputs(writer, tentacle_name=None):
    """
    Delete all user inputs. Only delete user inputs associated to tentacle_name if tentacle_name is given
    """
    if tentacle_name is None:
        await writer.delete_all(enums.DBTables.INPUTS.value)
    else:
        query = {"tentacle": tentacle_name}
        await writer.delete(enums.DBTables.INPUTS.value, query)


def _find_parent_config_node(tentacle_config, parent_input_name, array_indexes):
    """
    :return: the found parent node from tentacles_config
    """
    if parent_input_name is not None:
        found, nested_parent = dict_util.find_nested_value(
            tentacle_config,
            sanitize_user_input_name(parent_input_name),
            list_indexes=array_indexes,
        )
        if found and isinstance(nested_parent, dict):
            return nested_parent
        if found and isinstance(nested_parent, list) and array_indexes:
            return nested_parent[array_indexes[-1]]
        # non dict or list with array_indexes nested parents are not supported
        return None
    return tentacle_config



# pylint: disable=C0415, W0603, W1508, R0913
#  Drakkar-Software OctoBot-Commons
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import contextlib
import logging
import typing

import octobot_commons.constants as constants
import octobot_commons.timestamp_util as timestamp_util
import octobot_commons.html_util as html_util

LOG_DATABASE = "log_db"
LOG_NEW_ERRORS_COUNT = "log_new_errors_count"

BACKTESTING_NEW_ERRORS_COUNT: str = "log_backtesting_errors_count"

logs_database = {
    LOG_DATABASE: [],
    LOG_NEW_ERRORS_COUNT: 0,
    BACKTESTING_NEW_ERRORS_COUNT: 0,
}

error_notifier_callbacks = []

LOGS_MAX_COUNT = 1000

STORED_LOG_MIN_LEVEL = logging.WARNING
ENABLE_WEB_INTERFACE_LOGS = True
ERROR_PUBLICATION_ENABLED = True
SHOULD_PUBLISH_LOGS_WHEN_RE_ENABLED = False


def _default_callback(*_, **__):
    pass


_ERROR_CALLBACK = _default_callback
_LOG_CALLBACK: typing.Union[None, typing.Callable[[str], str]] = None


def set_global_logger_level(level, handler_levels=None) -> None:
    """
    Set the global logger level
    :param level: the level to set
    """
    logger = logging.getLogger()
    logger.setLevel(level)
    levels = handler_levels or [level] * len(logger.handlers)
    for handler, updated_level in zip(logger.handlers, levels):
        handler.setLevel(updated_level)


def get_global_logger_level() -> object:
    """
    Return the global logger level
    :return: the global logger level
    """
    return logging.getLogger().getEffectiveLevel()


@contextlib.contextmanager
def temporary_log_level(level):
    """
    Sets the log level to the given level inside this context
    """
    previous_level = get_global_logger_level()
    try:
        set_global_logger_level(level)
        yield
    finally:
        set_global_logger_level(previous_level)


def get_logger_level_per_handler() -> list:
    """
    Return the global logger level
    :return: order handles logging levels
    """
    return [handler.level for handler in logging.getLogger().handlers]


def get_logger(logger_name="Anonymous"):
    """
    Return the logger from the logger_name
    :param logger_name: the logger name
    :return: the logger from the logger name
    """
    return BotLogger(logger_name)


def set_logging_level(logger_names, level) -> None:
    """
    Set the logging level for the logger names
    :param logger_names: the logger names
    :param level: the level to set
    """
    for name in logger_names:
        logging.getLogger(name).setLevel(level)


def add_log(level, source, message, keep_log=True, call_notifiers=True):
    """
    Add a log to the log database
    :param level: the log level
    :param source: the log source
    :param message: the log message
    :param keep_log: if the log should be stored
    :param call_notifiers: if the log should trigger the notifiers
    """
    if keep_log:
        logs_database[LOG_DATABASE].append(
            {
                "Time": timestamp_util.get_now_time(),
                "Level": logging.getLevelName(level),
                "Source": str(source),
                "Message": message,
            }
        )
        if len(logs_database[LOG_DATABASE]) > LOGS_MAX_COUNT:
            logs_database[LOG_DATABASE].pop(0)
        # do not count this error if keep_log is False
        if level >= logging.ERROR:
            logs_database[LOG_NEW_ERRORS_COUNT] += 1
            logs_database[BACKTESTING_NEW_ERRORS_COUNT] += 1
    if call_notifiers:
        for callback in error_notifier_callbacks:
            callback()


def get_errors_count(counter=LOG_NEW_ERRORS_COUNT):
    """
    Return the error count according to the specified counter
    :param counter: the counter to use
    :return: the error count
    """
    return logs_database[counter]


def reset_errors_count(counter=LOG_NEW_ERRORS_COUNT):
    """
    Reset the specified counter error count
    :param counter: the counter to use
    """
    logs_database[counter] = 0


def register_error_notifier(callback):
    """
    Register an error notifier
    :param callback: the callback to call when the notifier is triggered
    """
    error_notifier_callbacks.append(callback)


class BotLogger:
    """
    The bot logger that manage all OctoBot's logs
    """

    def __init__(self, logger_name):
        self.logger_name = logger_name
        self.logger = logging.getLogger(logger_name)

    def debug(self, message: str, *args, **kwargs) -> None:
        """
        Called for a debug log
        :param message: the log message
        """
        message = self._process_log_callback(message)
        self.logger.debug(message, *args, **kwargs)
        self._publish_log_if_necessary(message, logging.DEBUG)

    def info(self, message: str, *args, **kwargs) -> None:
        """
        Called for an info log
        :param message: the log message
        """
        message = self._process_log_callback(message)
        self.logger.info(message, *args, **kwargs)
        self._publish_log_if_necessary(message, logging.INFO)

    def warning(self, message: str, *args, **kwargs) -> None:
        """
        Called for a warning log
        :param message: the log message
        """
        message = self._process_log_callback(message)
        self.logger.warning(message, *args, **kwargs)
        self._publish_log_if_necessary(message, logging.WARNING)

    def error(self, message: str, *args, skip_post_callback=False, **kwargs) -> None:
        """
        Called for an error log
        :param message: the log message
        :param skip_post_callback: when True, the error callback wont be called
        """
        message = self._process_log_callback(message)
        self.logger.error(message, *args, **kwargs)
        self._publish_log_if_necessary(message, logging.ERROR)
        self._post_callback_if_necessary(None, message, skip_post_callback)

    def exception(
        self,
        exception: Exception,
        publish_error_if_necessary: bool = True,
        error_message: str = None,
        include_exception_name: bool = True,
        skip_post_callback: bool = False,
        **kwargs,
    ) -> None:
        """
        Called for an exception log
        :param exception: the log exception
        :param publish_error_if_necessary: if the error should be published
        :param error_message: the log message
        :param include_exception_name: when True adds the __class__.__name__ of the exception at the end of the message
        :param skip_post_callback: when True, the error callback won't be called
        """
        extra = kwargs.get("extra", {})
        origin_error_message = error_message
        error_message = (
            self._process_log_callback(error_message)
            if error_message
            else error_message
        )
        extra[constants.EXCEPTION_DESC] = error_message
        html_util.summarize_exception_html_cause_if_relevant(exception)
        self.logger.exception(exception, extra=extra, **kwargs)
        if publish_error_if_necessary:
            message = origin_error_message
            if message is None:
                message = (
                    str(exception) if str(exception) else exception.__class__.__name__
                )
            elif include_exception_name:
                message = f"{message} (error: {exception.__class__.__name__})"
            self.error(
                message,
                skip_post_callback=True,
                extra={constants.IS_EXCEPTION_DESC: True},
            )
        self._post_callback_if_necessary(exception, error_message, skip_post_callback)

    def critical(self, message: str, *args, **kwargs) -> None:
        """
        Called for a critical log
        :param message: the log message
        """
        message = self._process_log_callback(message)
        self.logger.critical(message, *args, **kwargs)
        self._publish_log_if_necessary(message, logging.CRITICAL)

    def fatal(self, message: str, *args, **kwargs) -> None:
        """
        Called for a fatal log
        :param message: the log message
        """
        message = self._process_log_callback(message)
        self.logger.fatal(message, *args, **kwargs)
        self._publish_log_if_necessary(message, logging.FATAL)

    def disable(self, disabled):
        """
        Used to disable or enable this logger
        :param disabled: True to disable
        """
        self.logger.disabled = disabled

    def _process_log_callback(self, message: str) -> str:
        if _LOG_CALLBACK is None:
            return message
        return _LOG_CALLBACK(message)

    def _publish_log_if_necessary(self, message, level) -> None:
        """
        Publish the log message if necessary
        :param message: the log message
        :param level: the log level
        """
        if (
            ENABLE_WEB_INTERFACE_LOGS
            and STORED_LOG_MIN_LEVEL <= level
            and get_global_logger_level() <= level
        ):
            self._web_interface_publish_log(message, level)
            if not ERROR_PUBLICATION_ENABLED and logging.ERROR <= level:
                global SHOULD_PUBLISH_LOGS_WHEN_RE_ENABLED
                SHOULD_PUBLISH_LOGS_WHEN_RE_ENABLED = True

    def _web_interface_publish_log(self, message, level) -> None:
        """
        Publish log to web interface
        :param message: the log message
        :param level: the log level
        """
        add_log(
            level,
            self.logger_name,
            message,
            call_notifiers=ERROR_PUBLICATION_ENABLED,
        )

    @staticmethod
    def register_error_callback(callback):
        """
        :param callback: the callback to be called upon errors and exceptions
        Register callback as the ERROR_CALLBACK
        """
        global _ERROR_CALLBACK
        _ERROR_CALLBACK = callback

    @staticmethod
    def _post_callback_if_necessary(exception, error_message, skip_post_callback):
        if not skip_post_callback:
            _ERROR_CALLBACK(exception, error_message)


def register_log_callback(callback: typing.Union[None, typing.Callable[[str], str]]):
    """
    :param callback: the callback to be called upon any log of any level
    Register callback as the _LOG_CALLBACK
    """
    global _LOG_CALLBACK
    _LOG_CALLBACK = callback


def get_backtesting_errors_count() -> int:
    """
    Get backtesting errors count
    :return: the backtesting errors count
    """
    return get_errors_count(BACKTESTING_NEW_ERRORS_COUNT)


def reset_backtesting_errors() -> None:
    """
    Reset the backtesting errors count
    """
    reset_errors_count(BACKTESTING_NEW_ERRORS_COUNT)


def set_error_publication_enabled(enabled) -> None:
    """
    Set the error publication enabling
    :param enabled: if the error publication is enabled
    """
    global ERROR_PUBLICATION_ENABLED
    global SHOULD_PUBLISH_LOGS_WHEN_RE_ENABLED
    ERROR_PUBLICATION_ENABLED = enabled
    if enabled and SHOULD_PUBLISH_LOGS_WHEN_RE_ENABLED:
        add_log(logging.ERROR, None, None, keep_log=False, call_notifiers=True)
    else:
        SHOULD_PUBLISH_LOGS_WHEN_RE_ENABLED = False


def set_enable_web_interface_logs(enabled):
    """
    Disable or enable errors storage in web interface
    """
    global ENABLE_WEB_INTERFACE_LOGS
    ENABLE_WEB_INTERFACE_LOGS = enabled



# pylint: disable=W0718, W1203
# Drakkar-Software OctoBot-Commons
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import html.parser
import collections
import typing

# avoid cyclic imports with commons_logging
import logging


_IGNORED_ELEMENTS = ["script", "button"]
DEFAULT_ELEMENT_TEXT_MAX_SIZE = 100
MAX_RECURSIVE_EXCEPTION_CAUSES_DEPTH = 20


def summarize_page_content(
    html_content: str, max_element_text_size: int = DEFAULT_ELEMENT_TEXT_MAX_SIZE
) -> list[tuple[str, str]]:
    """
    :return: a list of (tag, content) tuples representing a html page's useful content
    """
    parser = _SummarizerHTMLParser(max_element_text_size)
    parser.feed(html_content)
    return parser.summary


def pretty_print_summary(summary: list[tuple[str, str]]) -> str:
    """
    :return: a str representing the summary of the page
    """
    return "; ".join(f"{element[0]}<{element[1]}>" for element in summary)


def is_html_content(html_content: str) -> bool:
    """
    :return: True if the given html_content looks like html
    """
    return "</html>" in html_content


def get_html_summary_if_relevant(
    html_content: typing.Any, max_element_text_size: int = DEFAULT_ELEMENT_TEXT_MAX_SIZE
) -> typing.Any:
    """
    :return: the str summary of the given html_content if
    it is html, the given html_content otherwise
    """
    try:
        str_html_content = (
            html_content if isinstance(html_content, str) else str(html_content)
        )
        if is_html_content(str_html_content):
            return pretty_print_summary(
                summarize_page_content(
                    str_html_content, max_element_text_size=max_element_text_size
                )
            )
        return str_html_content
    except BaseException as err:
        logging.getLogger("html_util").error(
            f"Error when parsing html_content '{html_content}', "
            f"error: {err} ({err.__class__.__name__})"
        )
    return html_content


def summarize_exception_html_cause_if_relevant(exception: BaseException, depth=0):
    """
    Recursively updates args and the __cause__ attribute
    of the exception to summarize html content if any
    """
    try:
        # Optimistic consideration of attributes being available: should always be the case.
        # However, if this is not the case, just catch the error not forward it
        if exception is not None:
            exception.args = tuple(
                get_html_summary_if_relevant(arg) for arg in exception.args
            )
            # condition should not be necessary but still make sure to avoid infinite recursive loops
            if depth < MAX_RECURSIVE_EXCEPTION_CAUSES_DEPTH:
                # recursive call to make sure nested causes are also summarized
                summarize_exception_html_cause_if_relevant(
                    exception.__cause__, depth=depth + 1
                )
    except BaseException:
        # Can't format html, nothing to do: just stop processing
        pass


class _SummarizerHTMLParser(html.parser.HTMLParser):
    """
    Walks through the give html document and stores its relevant into self.summary
    """

    # from https://docs.python.org/3/library/html.parser.html

    def __init__(self, max_element_text_size: int):
        super().__init__()
        self.summary: list[tuple[str, str]] = []

        self._path = collections.deque()
        self._max_element_text_size = max_element_text_size

    def handle_starttag(self, tag, attrs):
        self._path.append(tag)

    def handle_endtag(self, tag):
        self._path.pop()

    def handle_data(self, data):
        cleared_data = data.strip()
        if len(cleared_data) > self._max_element_text_size:
            cleared_data = f"{cleared_data[:self._max_element_text_size]}[...]"
        if len(cleared_data) > 1:
            try:
                element_name = self._path[-1]
                if element_name not in _IGNORED_ELEMENTS:
                    self.summary.append((element_name, cleared_data))
            except IndexError:
                # before or after html content
                self.summary.append(("message", cleared_data))
            except BaseException as err:
                logging.getLogger(self.__class__.__name__).error(
                    f"Error when parsing element for: '{data}', "
                    f"error: {err} ({err.__class__.__name__})",
                )



#  Drakkar-Software OctoBot-Commons
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import typing

import octobot_commons.profiles.profile_data as profile_data_import
import octobot_commons.profiles.exchange_auth_data as exchange_auth_data_import


class TentaclesProfileDataAdapter:
    """
    Used to adapt the content of a ProfileData using the given TentaclesData
    """

    def __init__(
        self,
        tentacles_data: list[profile_data_import.TentaclesData],
        additional_data: dict,
        authenticator,
        auth_key: typing.Optional[str],
    ):
        self.tentacles_data: list[profile_data_import.TentaclesData] = tentacles_data
        self.additional_data: dict = additional_data
        self.authenticator = authenticator
        self.auth_key = auth_key

    async def adapt(
        self,
        profile_data: profile_data_import.ProfileData,
        auth_data: list[exchange_auth_data_import.ExchangeAuthData],
    ) -> None:
        """
        Use self.tentacles_data to adapt the given profile_data
        """
        raise NotImplementedError("adapt is not implemented")

    @classmethod
    def get_tentacle_name(cls) -> str:
        """
        :return: the name of the adapter
        """
        raise NotImplementedError("get_tentacle_name is not implemented")



# pylint: disable=W0718
#  Drakkar-Software OctoBot-Commons
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import ssl
import time
import urllib.parse
import dataclasses

import contextlib
import aiohttp
import aiohttp.typedefs
import certifi

import octobot_commons.logging
import octobot_commons.constants


async def download_stream_file(
    output_file,
    file_url,
    aiohttp_session,
    data_chunk_size=5 * 2**20,
    is_aiofiles_output_file=False,
) -> str:
    """
    Download a big file with an aiohttp session
    :param output_file: the output file
    :param file_url: the file to be downloaded url
    :param aiohttp_session: the aiohttp session
    :param data_chunk_size: default value is 5*2**20 (5MB)
    :param is_aiofiles_output_file: When True, output_file.write will be awaited (when it's an aiofiles instance)
    :return downloaded file last_modified if given as response header
    """
    last_modified = None
    async with aiohttp_session.get(file_url) as resp:
        if resp.status != 200:
            try:
                text = await resp.text()
            except BaseException as err:
                text = f"error when reading resp text: {err}"
            raise RuntimeError(
                f"Failed to download file at url : {file_url} (status: {resp.status}, text: {text})"
            )
        while True:
            last_modified = resp.headers.get("Last-Modified", "unknown")
            chunk = await resp.content.read(data_chunk_size)
            if not chunk:
                # resp.content.read returns an empty chunk when completed
                break
            if is_aiofiles_output_file:
                await output_file.write(chunk)
            else:
                output_file.write(chunk)
    return last_modified


async def _check_local_certificates_availability(
    session: aiohttp.ClientSession, test_url: str
):
    try:
        # try fetching https://tentacles.octobot.online/ using local certificates
        async with session.get(test_url) as resp:
            if resp.status >= 400:
                octobot_commons.logging.get_logger(__name__).error(
                    f"Error when checking ssl certificates: fetching {test_url} returned {resp.status}. "
                    f"Considering certificates as valid."
                )
            return True
    except aiohttp.ClientConnectorCertificateError:
        return False
    except Exception as err:
        octobot_commons.logging.get_logger(__name__).info(
            f"Impossible to check ssl certificate: {err}. Consider valid"
        )
        return True


def _get_certify_aiohttp_client_session() -> aiohttp.ClientSession:
    # from https://docs.aiohttp.org/en/stable/client_advanced.html#example-use-certifi
    ssl_context = ssl.create_default_context(cafile=certifi.where())
    return aiohttp.ClientSession(connector=aiohttp.TCPConnector(ssl=ssl_context))


async def get_ssl_fallback_aiohttp_client_session(
    test_url: str,
) -> aiohttp.ClientSession:
    """
    :return: an aiohttp.ClientSession using certifi ssl certificates if necessary
    """
    base_session = aiohttp.ClientSession()
    if (
        not octobot_commons.constants.ENABLE_CERTIFI_SSL_CERTIFICATES
        or await _check_local_certificates_availability(base_session, test_url)
    ):
        return base_session
    try:
        await base_session.close()
    except Exception as err:
        octobot_commons.logging.get_logger(__name__).exception(
            err, True, f"Error when closing test session: {err}"
        )
    # use custom SSL certificates session
    fallback_session = _get_certify_aiohttp_client_session()
    octobot_commons.logging.get_logger(__name__).info(
        "Falling back to certifi configured aiohttp connector."
    )
    return fallback_session


@contextlib.asynccontextmanager
async def ssl_fallback_aiohttp_client_session(test_url: str):
    """
    yields an aiohttp.ClientSession using certifi ssl certificates if necessary
    """
    session = None
    try:
        session = await get_ssl_fallback_aiohttp_client_session(test_url)
        yield session
    finally:
        if session is not None:
            await session.close()


@contextlib.asynccontextmanager
async def certify_aiohttp_client_session():
    """
    yields an aiohttp.ClientSession always using certifi ssl certificates
    """
    session = None
    try:
        octobot_commons.logging.get_logger(__name__).debug(
            "Using certifi configured aiohttp connector."
        )
        session = _get_certify_aiohttp_client_session()
        yield session
    finally:
        if session is not None:
            await session.close()


@dataclasses.dataclass
class RequestCounter:
    """
    A counter with an identifier: counts requests and logs when its period is over
    """

    name: str
    period: float
    last_period_start: float = 0
    paths: dict[str, float] = dataclasses.field(default_factory=dict)

    def account_for(
        self, method: str, str_or_url: aiohttp.typedefs.StrOrURL, timestamp: float
    ):
        """
        Account for a request, log if period is over
        """
        if timestamp - self.last_period_start > self.period:
            if self.last_period_start != 0:
                self._log_stats()
            self._clear()
            self.last_period_start = timestamp - timestamp % self.period
        url = str_or_url if isinstance(str_or_url, str) else str_or_url.human_repr()
        path = f"[{method}] { urllib.parse.urlparse(url).path}"
        self.paths[path] = self.paths.get(path, 0) + 1

    def _clear(self):
        self.paths.clear()

    def _log_stats(self):
        total_req = sum(value for value in self.paths.values())
        hours = self.period / octobot_commons.constants.HOURS_TO_SECONDS
        days = self.period / octobot_commons.constants.DAYS_TO_SECONDS
        identifier = (
            f"{days} days"
            if days > 1
            else f"{hours} hours"
            if hours > 1
            else f"{self.period} seconds"
        )
        octobot_commons.logging.get_logger(self.__class__.__name__).info(
            f"[{self.name}]: {total_req} requests over {identifier}: {self.paths}"
        )


class CounterClientSession(aiohttp.ClientSession):
    """
    Extends aiohttp.ClientSession to periodically log requests statistics
    """

    def __init__(self, identifier: str, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.per_min: RequestCounter = RequestCounter(
            identifier, octobot_commons.constants.MINUTE_TO_SECONDS
        )
        self.per_hour: RequestCounter = RequestCounter(
            identifier, octobot_commons.constants.HOURS_TO_SECONDS
        )
        self.per_day: RequestCounter = RequestCounter(
            identifier, octobot_commons.constants.DAYS_TO_SECONDS
        )

    async def _request(
        self, method: str, str_or_url: aiohttp.typedefs.StrOrURL, *args, **kwargs
    ) -> aiohttp.ClientResponse:
        self._account_for(method, str_or_url, time.time())
        if kwargs:
            return await super()._request(method, str_or_url, *args, **kwargs)
        return await super()._request(method, str_or_url)

    def _account_for(
        self, method: str, str_or_url: aiohttp.typedefs.StrOrURL, timestamp: float
    ):
        try:
            self.per_min.account_for(method, str_or_url, timestamp)
            self.per_hour.account_for(method, str_or_url, timestamp)
            self.per_day.account_for(method, str_or_url, timestamp)
        except BaseException as err:
            # never raise or the subsequent request is blocked
            octobot_commons.logging.get_logger(__name__).exception(
                err, True, "Error when accounting for request: {err}"
            )
